from helpers.conversation import load_conversations, summarise_conversation
from helpers.clusters import cluster_summaries, reduce_clusters, embed_summaries
from asyncio import run, Semaphore
import instructor
import google.generativeai as genai
from tqdm.asyncio import tqdm_asyncio as asyncio
from rich import print
import json
import os
from openai import AsyncOpenAI

from helpers.types import Cluster


async def generate_clusters(
    SUMMARIES_PER_CLUSTER=20,
    CHILD_CLUSTERS_PER_CLUSTER=10,
    MAX_FINAL_CLUSTERS=10,
    start_step=None,
):
    conversations = load_conversations("conversations.json")

    # Step 1: Generate or load summaries
    if start_step == "summarize" or not os.path.exists("checkpoints/summaries.json"):
        client = instructor.from_gemini(
            genai.GenerativeModel("gemini-1.5-flash-latest"),
            use_async=True,
        )
        sem = Semaphore(50)

        summaries = await asyncio.gather(
            *[
                summarise_conversation(client, sem, conversation)
                for conversation in conversations
            ]
        )

        # Save summaries
        with open("checkpoints/summaries.json", "w") as f:
            for summary in summaries:
                f.write(json.dumps(summary) + "\n")
        print("Generated and saved summaries")
    else:
        with open("checkpoints/summaries.json", "r") as f:
            summaries = [json.loads(line) for line in f]
        print("Loaded existing summaries")

    # Step 2: Generate or load embeddings
    if start_step in ["summarize", "embed"] or not os.path.exists(
        "checkpoints/embedded_summaries.json"
    ):
        oai_client = AsyncOpenAI()
        sem = Semaphore(10)
        await embed_summaries(oai_client, sem, summaries)
        print("Generated and saved embeddings")

    # Step 3: Generate or load base clusters
    if start_step in ["summarize", "embed", "cluster"] or not os.path.exists(
        "checkpoints/base_clusters.json"
    ):
        client = instructor.from_gemini(
            genai.GenerativeModel("gemini-1.5-flash-latest"),
            use_async=True,
        )
        sem = Semaphore(10)

        base_clusters = await cluster_summaries(
            client, sem, summaries, SUMMARIES_PER_CLUSTER
        )

        # Save base clusters
        with open("checkpoints/base_clusters.json", "w") as f:
            for cluster in base_clusters:
                f.write(cluster.model_dump_json() + "\n")
        print(f"Generated and saved {len(base_clusters)} base clusters")
    else:
        with open("checkpoints/base_clusters.json", "r") as f:
            base_clusters = [Cluster.model_validate_json(line) for line in f]
        print("Loaded existing base clusters")

    # Create checkpoints directory if it doesn't exist
    os.makedirs("checkpoints", exist_ok=True)

    clusters = base_clusters
    root_clusters = base_clusters
    iteration = 0

    while len(root_clusters) >= MAX_FINAL_CLUSTERS:
        print(f"\nIteration {iteration}")
        print(f"Starting with {len(root_clusters)} root clusters")

        # Get both updated original clusters and new parent clusters
        reduced_clusters = await reduce_clusters(
            client, sem, root_clusters, CHILD_CLUSTERS_PER_CLUSTER
        )

        # Recalculate root clusters for next iteration
        root_clusters = [c for c in reduced_clusters if not c.parent_id]

        # Remove the outdated versions of clusters that were just reduced
        old_cluster_ids = {rc.id for rc in root_clusters}
        clusters = [c for c in clusters if c.id not in old_cluster_ids]

        # Add both the updated original clusters and new parent clusters
        clusters.extend(reduced_clusters)

        print(f"Reduced to {len(root_clusters)} root clusters")
        print(f"Total clusters after reduction: {len(clusters)}")

        iteration += 1

    print(f"\nFinal reduction complete:")
    print(f"Root clusters: {len(root_clusters)}")
    print(f"Total clusters (including children): {len(clusters)}")

    # Save final clusters
    with open("clusters.json", "w") as f:
        for cluster in clusters:
            f.write(cluster.model_dump_json() + "\n")


if __name__ == "__main__":
    run(generate_clusters())



from .types import Cluster, GeneratedCluster
from openai import AsyncOpenAI
from asyncio import Semaphore
from tqdm.asyncio import tqdm_asyncio as asyncio
import math
import re
import random
import numpy as np
from sklearn.cluster import KMeans
from pydantic import BaseModel, field_validator, ValidationInfo
from instructor import from_gemini, AsyncInstructor
import google.generativeai as genai


class CandidateClusters(BaseModel):
    candidate_cluster_names: list[str]

    @field_validator("candidate_cluster_names")
    def validate_candidate_cluster_names(
        cls, v: list[str], info: ValidationInfo
    ) -> list[str]:
        if len(v) == 0:
            raise ValueError("Candidate cluster names must be a non-empty list")

        v = [label.strip() for label in v]
        v = [label[:-1] if label.endswith(".") else label for label in v]

        return [re.sub(r"\\{1,}", "", label.replace('"', "")) for label in v]


class ClusterLabel(BaseModel):
    higher_level_cluster: str

    @field_validator("higher_level_cluster")
    def validate_higher_level_cluster(cls, v: str, info: ValidationInfo) -> str:
        if v not in info.context["candidate_clusters"]:
            raise ValueError(
                f"""
                Invalid higher-level cluster: |{v}|
                
                Valid clusters are:
                {', '.join(f'|{c}|' for c in info.context['candidate_clusters'])}
                """
            )
        return v


class ClusterModel:
    def __init__(
        self,
        embed_model: AsyncOpenAI = AsyncOpenAI(),
        cluster_model: AsyncInstructor = from_gemini(
            genai.GenerativeModel("gemini-1.5-flash-latest"), use_async=True
        ),
        max_concurrent_calls: int = 40,
        base_cluster_per_cluster: int = 20,
    ) -> None:
        self.sem = Semaphore(max_concurrent_calls)
        self.embed_model = embed_model
        self.cluster_model = cluster_model
        self.base_cluster_per_cluster = base_cluster_per_cluster

    async def embed_cluster(self, cluster: Cluster) -> Cluster:
        async with self.sem:
            embedding = await self.embed_model.embeddings.create(
                input=f"""
                Name: {cluster.name}
                Description: {cluster.description}
                """,
                model="text-embedding-3-small",
            )
            return {
                "item": cluster,
                "embedding": embedding.data[0].embedding,
            }

    def cluster_base_clusters(
        self, clusters: list[Cluster]
    ) -> dict[int, list[Cluster]]:
        n_clusters = math.ceil(len(clusters) / self.base_cluster_per_cluster)

        if n_clusters == 1:
            return {0: [item["item"] for item in clusters]}

        embeddings = [item["embedding"] for item in clusters]
        X = np.array(embeddings)

        kmeans = KMeans(n_clusters=n_clusters)
        cluster_labels = kmeans.fit_predict(X)

        group_to_clusters = {}
        for i, item in enumerate(clusters):
            cluster_id = int(cluster_labels[i])
            if cluster_id not in group_to_clusters:
                group_to_clusters[cluster_id] = []

            group_to_clusters[cluster_id].append(item["item"])

        return group_to_clusters

    async def generate_candidate_cluster_names(
        self, clusters: list[Cluster]
    ) -> list[str]:
        async with self.sem:
            resp = await self.cluster_model.chat.completions.create(
                messages=[
                    {
                        "role": "system",
                        "content": """
                You are tasked with creating higher-level cluster names based on a given list of clusters and their descriptions. Your goal is to come up with broader categories that could encompass one or more of the provided clusters

                First, review the list of clusters and their descriptions:
                <cluster_list>
                    {% for cluster in clusters %}
                    <cluster>{{ cluster.name }}: {{ cluster.description }}</cluster>
                    {% endfor %}
                </cluster_list>

                Your task is to create at most {{ desired_number }} higher-level cluster names that could potentially include one or more of the provided clusters. These higher-level clusters should represent broader categories or themes that emerge from the given clusters, while remaining as specific as possible. If there are many clusters with a specific theme, ensure that the higher-level cluster name remains the maximum level of specificity. You are helping to organize user behavior data in order to improve safety, monitoring, and observability. You can generate less than {{ desired_number }} names if you feel that fewer are appropriate and accurately capture the clusters. 

                Guidelines for creating higher-level clusters names
                1. Analyze the themes, topics or characteristics common to multiple clusters.
                2. Create names that are specific enough to be meaningful but but not so specific that they can’t meaningfully represent many different clusters. Avoid overly general or vague terms, and do not hesitate to describe socially harmful or sensitive topics (in fact, clusters that clearly describe harmful behavior are slightly preferred); specificity is necessary for observability and enforcement.
                3. Ensure that the higher-level cluster names are distinct from one another.
                4. Use clear, concise, and descriptive language for the cluster names. Assume neither good nor bad faith for the content in the clusters.

                Think about the relationships between the given clusters and potential overarching themes. 
                
                Focus on creating meaningful, distinct and precise ( but not overly specific ) higher-level cluster names that could encompass multiple sub-clusters.
                """,
                    }
                ],
                response_model=CandidateClusters,
                context={
                    "clusters": clusters,
                    "desired_number": math.ceil(len(clusters) / 2),
                },
                max_retries=3,
            )
            return resp.candidate_cluster_names

    async def label_cluster(
        self,
        cluster: Cluster,
        candidate_clusters: list[str],
    ) -> dict:
        async with self.sem:
            candidate_clusters = candidate_clusters.copy()
            random.shuffle(candidate_clusters)
            label = await self.cluster_model.chat.completions.create(
                messages=[
                    {
                        "role": "system",
                        "content": """
                        You are tasked with categorizing a specific cluster into one of the provided higher-level clusters for observability, monitoring, and content moderation. Your goal is to determine which higher-level cluster best fits the given specific cluster based on its name and description.

                        First, here are the ONLY valid higher-level clusters you may select from:
                        <higher_level_clusters>
                        {% for cluster in candidate_clusters %}
                        <higher_level_cluster>{{ cluster }}</higher_level_cluster>
                        {% endfor %}
                        </higher_level_clusters>

                        Here is the specific cluster to categorize:
                        <specific_cluster>
                        Name: {{ cluster["name"] }}
                        Description: {{ cluster["description"] }}
                        </specific_cluster>

                        RULES:
                        1. You MUST select EXACTLY ONE higher-level cluster from the provided list
                        2. You MUST output the higher-level cluster name EXACTLY as written - no modifications allowed
                        3. You MUST NOT create new cluster names or combinations
                        4. You MUST NOT output any additional text or explanations
                        5. You MUST NOT use partial matches or approximate names

                        CLASSIFICATION PROCESS:
                        1. First, record the exact list of valid higher-level clusters
                        2. Read the specific cluster's name and description carefully
                        3. Compare the specific cluster's key characteristics against each valid higher-level cluster
                        4. Select the single most appropriate higher-level cluster that encompasses the specific cluster
                        5. Verify your selected cluster exactly matches one from the valid list
                        6. Output ONLY the selected higher-level cluster name, exactly as it appeared in the valid list

                        Here is the specific cluster to categorize:

                        <specific_cluster>
                        Name: {{ cluster["name"] }}
                        Description: {{ cluster["description"] }}
                        </specific_cluster>
                        
                        Based on this information, determine the most appropriate higher-level cluster and provide your answer as instructed.
                        """,
                    }
                ],
                context={
                    "cluster": cluster,
                    "candidate_clusters": candidate_clusters,
                },
                response_model=ClusterLabel,
                max_retries=3,
            )
        return {
            "cluster": cluster,
            "higher_level_cluster": label.higher_level_cluster,
        }

    async def rename_cluster_group(self, clusters: list[Cluster]) -> Cluster:
        async with self.sem:
            resp = await self.cluster_model.chat.completions.create(
                messages=[
                    {
                        "role": "system",
                        "content": """
                        You are tasked with summarizing a group of related cluster names into a short, precise, and accurate overall description and name. Your goal is to create a concise summary that captures the essence of these clusters

                        The cluster name should be at most ten words long (perhaps less) and be specific but also reflective of most of the clusters that comprise them. The cluster name should be a sentence in the imperative that captures the user’s request. For example, ‘Brainstorm ideas for a birthday party’ or ‘Help me find a new job are good examples.

                        Do not shy away from describing socially harmful or sensitive topics. For instance, "Write fantasy sexual roleplay with octopi and monsters", "Generate blog spam for gambling websites", or "Assist with high school math homework" would be better than "Write erotic content" or "Help with homework". Be as descriptive as possible and assume neither good nor bad faith. Do not hesitate to identify and describe socially harmful or sensitive topics specifically; specificity is necessary for monitoring.

                        The cluster description should be a clear, precise, two-sentence description in the past tense. This description should be specific to this cluster. Make sure that you've captured the most important details of the cluster.
                        
                        Below are the related cluster names
                        <clusters>
                            {% for cluster in clusters %}
                                <cluster>{{ cluster.name }}: {{ cluster.description }}</cluster>
                            {% endfor %}
                        </clusters>

                        Ensure your summary and name accurately represent the clusters and are specific to the clusters.
                        """,
                    }
                ],
                context={"clusters": clusters},
                response_model=GeneratedCluster,
            )

        res = []

        new_cluster = Cluster(
            name=resp.name,
            description=resp.summary,
            chat_ids=[chat_id for cluster in clusters for chat_id in cluster.chat_ids],
            parent_id=None,
        )

        res.append(new_cluster)

        for cluster in clusters:
            res.append(
                Cluster(
                    id=cluster.id,
                    name=cluster.name,
                    description=cluster.description,
                    chat_ids=cluster.chat_ids,
                    parent_id=new_cluster.id,
                )
            )

        return res

    async def generate_higher_level_cluster(
        self, clusters: list[Cluster]
    ) -> list[Cluster]:
        """
        This takes in a list of clusters and generates a smaller list of clusters that are more general.

        We do so in a 3 step process

        1. First we generate a list of candidate cluster names
        2. Then we label each base cluster with a potential candidate cluster name
        3. Lastly, for our new clusters, we then generate a new re-labelled cluster with a new title and description that is more representative of these new clusters

        We do have a base case, in which that if a cluster contains a single item, it's just returned as is.
        """
        if len(clusters) == 1:
            return clusters

        candidates_cluster_names = await self.generate_candidate_cluster_names(clusters)

        labelled_clusters = await asyncio.gather(
            *[
                self.label_cluster(cluster, candidates_cluster_names)
                for cluster in clusters
            ],
            disable=True,
        )
        label_to_cluster = {}
        for labelled_cluster in labelled_clusters:
            if labelled_cluster["higher_level_cluster"] not in label_to_cluster:
                label_to_cluster[labelled_cluster["higher_level_cluster"]] = []

            label_to_cluster[labelled_cluster["higher_level_cluster"]].append(
                labelled_cluster["cluster"]
            )

        new_clusters = await asyncio.gather(
            *[
                self.rename_cluster_group(cluster)
                for cluster in label_to_cluster.values()
            ],
            disable=True,
        )

        final_clusters = []

        for clusters in new_clusters:
            final_clusters.extend(clusters)

        return final_clusters

    async def cluster_clusters(self, clusters: list[Cluster]) -> list[Cluster]:
        embedded_clusters = await asyncio.gather(
            *[self.embed_cluster(cluster) for cluster in clusters], disable=True
        )
        clusters = self.cluster_base_clusters(embedded_clusters)
        new_clusters = await asyncio.gather(
            *[
                self.generate_higher_level_cluster(cluster)
                for cluster in clusters.values()
            ],
            desc="Generating New Higher Level Clusters",
        )

        res = []
        for clusters in new_clusters:
            res.extend(clusters)

        return res



from pydantic import BaseModel, Field, computed_field
from datetime import datetime
import uuid
from typing import Union


class Message(BaseModel):
    created_at: datetime
    role: str
    content: str


class Conversation(BaseModel):
    chat_id: str
    created_at: datetime
    messages: list[Message]


class ConversationSummary(BaseModel):
    chat_id: str
    task_description: str
    user_request: str
    metadata: dict


class Cluster(BaseModel):
    id: str = Field(
        default_factory=lambda: uuid.uuid4().hex,
    )
    name: str
    description: str
    chat_ids: list[str]
    parent_id: Union[str, None]

    @computed_field
    def count(self) -> int:
        return len(self.chat_ids)


class GeneratedCluster(BaseModel):
    name: str
    summary: str


class CachedCluster(Cluster):
    umap_embedding: list[float]



from .cluster import ClusterModel
from .types import Cluster, ConversationSummary, Conversation, Message
from .llm_cluster import LLMCluster
from .summarise import SummariseBase

__all__ = [
    "ClusterModel",
    "Cluster",
    "ConversationSummary",
    "Conversation",
    "Message",
    "LLMCluster",
    "SummariseBase",
]



from asyncio import Semaphore
from kura.types import ConversationSummary, GeneratedCluster, Cluster
from instructor import from_gemini
import google.generativeai as genai
from openai import AsyncOpenAI
from tqdm.asyncio import tqdm_asyncio as asyncio
import numpy as np
from sklearn.cluster import KMeans
import math
import os
import json


class SummaryCluster:
    def __init__(
        self,
        cluster_model=from_gemini(
            genai.GenerativeModel("gemini-1.5-flash-latest"), use_async=True
        ),
        embedding_model=AsyncOpenAI(),
        max_concurrent_calls: int = 40,
        summaries_per_cluster: int = 10,
        checkpoint_dir: str = "checkpoints",
        checkpoint_file_name: str = "base_clusters.json",
    ):
        self.sem = Semaphore(max_concurrent_calls)
        self.cluster_model = cluster_model
        self.embedding_model = embedding_model
        self.summaries_per_cluster = summaries_per_cluster
        self.checkpoint_dir = checkpoint_dir
        self.checkpoint_file_name = checkpoint_file_name

    async def _embed(self, summary: ConversationSummary):
        async with self.sem:
            embedding = await self.embedding_model.embeddings.create(
                input=summary.user_request,
                model="text-embedding-3-small",
            )
            return {
                "item": summary,
                "embedding": embedding.data[0].embedding,
            }

    async def cluster_summaries(
        self, summaries: list[ConversationSummary]
    ) -> dict[int, list[ConversationSummary]]:
        n_clusters = math.ceil(len(summaries) / self.summaries_per_cluster)
        embeddings = [summary["embedding"] for summary in summaries]
        X = np.array(embeddings)

        kmeans = KMeans(n_clusters=n_clusters)
        cluster_labels = kmeans.fit_predict(X)

        group_to_clusters = {}
        for i, summary in enumerate(summaries):
            cluster_id = int(cluster_labels[i])
            if cluster_id not in group_to_clusters:
                group_to_clusters[cluster_id] = []
            group_to_clusters[cluster_id].append(summary["item"])

        return group_to_clusters

    def get_contrastive_examples(
        self,
        cluster_id: int,
        cluster_id_to_summaries: dict[int, list[ConversationSummary]],
        desired_count: int = 10,
    ):
        other_clusters = [c for c in cluster_id_to_summaries.keys() if c != cluster_id]
        all_examples = []
        for cluster in other_clusters:
            all_examples.extend(cluster_id_to_summaries[cluster])

        # If we don't have enough examples, return all of them
        if len(all_examples) <= desired_count:
            return all_examples

        # Otherwise sample without replacement
        return list(np.random.choice(all_examples, size=desired_count, replace=False))

    async def generate_base_cluster(
        self,
        cluster_id: int,
        cluster_id_to_summaries: dict[int, list[ConversationSummary]],
    ):
        async with self.sem:
            contrastive_examples = self.get_contrastive_examples(
                cluster_id, cluster_id_to_summaries
            )
            positive_examples = cluster_id_to_summaries[cluster_id]
            cluster = await self.cluster_model.chat.completions.create(
                messages=[
                    {
                        "role": "system",
                        "content": """
                    You are tasked with summarizing a group of related statements into a short, precise, and accurate description and name. Your goal is to create a concise summary that captures the essence of these statements and distinguishes them from other similar groups of statements.
                    
                    Summarize all the statements into a clear, precise, two-sentence description in the past tense. Your summary should be specific to this group and distinguish it from the contrastive answers of the other groups.

                    After creating the summary, generate a short name for the group of statements. This name should be at most ten words long (perhaps less) and be specific but also reflective of most of the statements (rather than reflecting only one or two). The name should distinguish this group from the contrastive examples. For instance, "Write fantasy sexual roleplay with octopi and monsters", "Generate blog spam for gambling websites", or "Assist with high school math homework" would be better and more actionable than general terms like "Write erotic content" or "Help with homework". Be as descriptive as possible and assume neither good nor bad faith. Do not hesitate to identify and describe socially harmful or sensitive topics specifically; specificity is necessary for monitoring
                    
                    The cluster name should be a sentence in the imperative that captures the user’s request. For example, ‘Brainstorm ideas for a birthday party’ or ‘Help me find a new job.”
                    """,
                    },
                    {
                        "role": "user",
                        "content": """
                    Here are the relevant statements

                    Below are the related statements:
                    <positive_examples>
                    {% for item in positive_examples %}
                        <positive_example>{{ item.user_request }}</positive_example>
                    {% endfor %}
                    </positive_examples>  
                    
                    For context, here are statements from nearby groups that are NOT part of the group you’re summarizing
                    
                    <contrastive_examples>
                    {% for item in contrastive_examples %}
                    <contrastive_example>{{ item.user_request }}</contrastive_example>
                    {% endfor %}
                    </contrastive_examples>
                    
                    Remember to analyze both the statements and the contrastive statements carefully to ensure your summary and name accurately represent the specific group while distinguishing it from others.
                    """,
                    },
                ],
                response_model=GeneratedCluster,
                context={
                    "positive_examples": positive_examples,
                    "contrastive_examples": contrastive_examples,
                },
            )

            return Cluster(
                name=cluster.name,
                description=cluster.summary,
                chat_ids=[item.chat_id for item in positive_examples],
                parent_id=None,
            )

    async def embed_summaries(self, summaries: list[ConversationSummary]):
        coros = [self._embed(summary) for summary in summaries]
        return await asyncio.gather(*coros, desc="Embedding Summaries")

    def save_checkpoint(self, data: list[ConversationSummary]):
        with open(
            os.path.join(self.checkpoint_dir, self.checkpoint_file_name), "w"
        ) as f:
            for item in data:
                f.write(item.model_dump_json() + "\n")

    def load_checkpoint(self) -> list[ConversationSummary]:
        with open(
            os.path.join(self.checkpoint_dir, self.checkpoint_file_name), "r"
        ) as f:
            return [Cluster(**json.loads(line)) for line in f]

    def save_checkpoint_clusters(self, clusters: list[Cluster]):
        with open(
            os.path.join(self.checkpoint_dir, self.checkpoint_file_name), "w"
        ) as f:
            for cluster in clusters:
                f.write(cluster.model_dump_json() + "\n")

    async def cluster_conversation_summaries(
        self, summaries: list[ConversationSummary]
    ) -> list[Cluster]:
        # Load Checkpoint By Default
        if os.path.exists(os.path.join(self.checkpoint_dir, self.checkpoint_file_name)):
            print("Loading Base Cluster Checkpoint")
            return self.load_checkpoint()

        embedded_summaries = await self.embed_summaries(summaries)
        cluster_id_to_summaries = await self.cluster_summaries(embedded_summaries)
        clusters = await asyncio.gather(
            *[
                self.generate_base_cluster(cluster_id, cluster_id_to_summaries)
                for cluster_id in cluster_id_to_summaries.keys()
            ],
            desc="Generating Base Clusters",
        )
        self.save_checkpoint(clusters)
        return clusters



from datetime import datetime, timedelta
import pandas as pd
from fasthtml.common import Script
import json
from typing import List
from helpers.types import Conversation


def generate_cumulative_chart(conversations: List[Conversation]) -> Script:
    """
    Generate a cumulative word count chart for human messages in conversations.
    Returns a Script element containing the Plotly visualization.
    """
    # Create DataFrame from messages
    messages_data = []
    for conv in conversations:
        for msg in conv.messages:
            if msg.role == "human":
                messages_data.append(
                    {
                        "datetime": pd.to_datetime(
                            msg.created_at.replace("Z", "+00:00")
                        ),
                        "words": len(msg.content.split()),
                    }
                )

    df = pd.DataFrame(messages_data)
    df["week_start"] = df["datetime"].dt.to_period("W-MON").dt.start_time

    # Calculate weekly and cumulative word counts
    weekly_df = df.groupby("week_start")["words"].sum().reset_index()
    weekly_df["cumulative_words"] = weekly_df["words"].cumsum()
    weekly_df["week_start"] = weekly_df["week_start"].dt.strftime("%Y-%m-%d")

    data = json.dumps(
        {
            "data": [
                {
                    "x": weekly_df["week_start"].tolist(),
                    "y": weekly_df["cumulative_words"].tolist(),
                    "type": "scatter",
                    "mode": "lines",
                    "name": "Cumulative Words in Human Messages",
                    "line": {"color": "red"},
                }
            ],
            "layout": {
                "title": "Cumulative Words in Human Messages by Week",
                "xaxis": {"title": "Week Starting"},
                "yaxis": {"title": "Total Words"},
                "showlegend": True,
                "hovermode": "closest",
            },
        }
    )

    return Script(f"var data = {data}; Plotly.newPlot('myDiv', data);")


def generate_messages_per_chat_chart(conversations):
    # Create DataFrame from messages
    messages_data = []
    for conv in conversations:
        for msg in conv.messages:
            if msg.role == "human":
                messages_data.append(
                    {
                        "datetime": pd.to_datetime(
                            msg.created_at.replace("Z", "+00:00")
                        ),
                        "chat_id": conv.chat_id,
                    }
                )

    df = pd.DataFrame(messages_data)
    df["week_start"] = df["datetime"].dt.to_period("W-MON").dt.start_time

    # Calculate messages and unique chats per week
    weekly_messages = df.groupby("week_start").size().reset_index(name="message_count")
    weekly_chats = (
        df.groupby("week_start")["chat_id"].nunique().reset_index(name="chat_count")
    )

    # Merge and calculate average
    weekly_df = pd.merge(weekly_messages, weekly_chats, on="week_start")
    weekly_df["avg_messages"] = weekly_df["message_count"] / weekly_df["chat_count"]
    weekly_df["week_start"] = weekly_df["week_start"].dt.strftime("%Y-%m-%d")

    data = json.dumps(
        {
            "data": [
                {
                    "x": weekly_df["week_start"].tolist(),
                    "y": weekly_df["avg_messages"].tolist(),
                    "type": "scatter",
                    "mode": "lines+markers",
                    "name": "Average Messages per Chat",
                    "line": {"color": "coral"},
                    "marker": {"size": 6},
                }
            ],
            "layout": {
                "title": "Average Number of Messages per Chat by Week",
                "xaxis": {"title": "Week Starting"},
                "yaxis": {"title": "Average Messages per Chat"},
                "showlegend": True,
                "hovermode": "closest",
            },
        }
    )

    return Script(
        f"var messages_data = {data}; Plotly.newPlot('messagesDiv', messages_data);"
    )


def generate_messages_per_week_chart(conversations: List[Conversation]):
    messages_data = []
    for conv in conversations:
        for msg in conv.messages:
            messages_data.append(
                {
                    "datetime": pd.to_datetime(msg.created_at.replace("Z", "+00:00")),
                    "chat_id": conv.chat_id,
                }
            )

    df = pd.DataFrame(messages_data)
    df["week_start"] = df["datetime"].dt.to_period("W-MON").dt.start_time

    # Calculate messages per week
    weekly_messages = df.groupby("week_start").size().reset_index(name="message_count")
    weekly_messages["week_start"] = weekly_messages["week_start"].dt.strftime(
        "%Y-%m-%d"
    )

    data = json.dumps(
        {
            "data": [
                {
                    "x": weekly_messages["week_start"].tolist(),
                    "y": weekly_messages["message_count"].tolist(),
                    "type": "bar",
                    "name": "Messages per Week",
                    "marker": {"color": "coral"},
                }
            ],
            "layout": {
                "title": "Messages per Week",
                "xaxis": {"title": "Week Starting"},
                "yaxis": {"title": "Number of Messages"},
                "showlegend": True,
                "hovermode": "closest",
                "bargap": 0.15,
            },
        }
    )

    return Script(
        f"var messages_per_week = {data}; Plotly.newPlot('messagesPerWeekDiv', messages_per_week.data, messages_per_week.layout);"
    )


def generate_new_chats_per_week_chart(conversations: List[Conversation]):
    # Calculate new chats per week
    chat_starts = pd.DataFrame(
        [
            {
                "datetime": pd.to_datetime(conv.created_at.replace("Z", "+00:00")),
                "chat_id": conv.chat_id,
            }
            for conv in conversations
        ]
    )
    chat_starts["week_start"] = (
        chat_starts["datetime"].dt.to_period("W-MON").dt.start_time
    )
    weekly_chats = (
        chat_starts.groupby("week_start").size().reset_index(name="chat_count")
    )
    weekly_chats["week_start"] = weekly_chats["week_start"].dt.strftime("%Y-%m-%d")

    data = json.dumps(
        {
            "data": [
                {
                    "x": weekly_chats["week_start"].tolist(),
                    "y": weekly_chats["chat_count"].tolist(),
                    "type": "bar",
                    "name": "New Chats per Week",
                    "marker": {"color": "skyblue"},
                }
            ],
            "layout": {
                "title": "New Chats Created per Week",
                "xaxis": {"title": "Week Starting"},
                "yaxis": {"title": "Number of New Chats"},
                "showlegend": True,
                "hovermode": "closest",
                "bargap": 0.15,
            },
        }
    )

    return Script(
        f"var chats_per_week = {data}; Plotly.newPlot('chatsPerWeekDiv', chats_per_week.data, chats_per_week.layout);"
    )



import json
from helpers.types import Conversation, Message, ConversationSummary
from typing import List
import instructor
from asyncio import Semaphore


def load_conversations(path: str) -> List[Conversation]:
    """
    This is a function that loads in the conversations from the given path and yields the user's conversations
    """
    with open(path, "r") as f:
        chat_obj = json.load(f)
        return [
            Conversation(
                chat_id=chat["uuid"],
                created_at=chat["created_at"],
                messages=[
                    Message(
                        created_at=message["created_at"],
                        role=message["sender"],
                        content="\n".join(
                            [
                                item["text"]
                                for item in message["content"]
                                if item["type"] == "text"
                            ]
                        ),
                    )
                    for message in chat["chat_messages"]
                ],
            )
            for chat in chat_obj
        ]


async def summarise_conversation(
    client: instructor.AsyncInstructor, sem: Semaphore, conversation: Conversation
) -> dict:
    """
    This is a function that given a user conversation will return a summary of the user request and the task description

    Task Description: The task is to do X
    User Request: The user's overall request for the assistant is to do Y

    """
    async with sem:
        resp = await client.chat.completions.create(
            messages=[
                {
                    "role": "system",
                    "content": """Generate a summary of the task that the user is asking and a description of the task that the assistant is trying to complete based off the following conversation.

                    Here are some examples of user requests
                    <examples>
                    - The user’s overall request for the assistant is to help implementing a React component to display a paginated list of users from a database.
                    - The user’s overall request for the assistant is to debug a memory leak in their Python data processing pipeline.
                    - The user’s overall request for the assistant is to design and architect a REST API for a social media application.
                    </examples>

                    Task descriptions should be concise and short. Here are some examples of task descriptions

                    <examples>
                    The task is to help build a frontend component with React and implement database integration
                    The task is to debug performance issues and optimize memory usage in Python code
                    The task is to design and architect a RESTful API following best practices
                    </examples>

                    Here is the conversation
                    {% for message in messages %}
                    <messages>
                    <message>{{message.role}}: {{message.content}} </message>
                    </messages>
                    {% endfor %}

                    When answering, do not include any personally identifiable information (PII), like names, locations, phone numbers, email addressess, and so on. When answering, do not include any proper nouns. Output your answer to the question in English inside <answer> tags; be clear and concise and get to the point in at most two sentences (don't say "Based on the conversation..." and avoid mentioning Claude/the chatbot). For example:

                    Remember that
                    - User requests and task descriptions should be concise and short. They should each be at most 1-2 sentences and at most 20 words.
                    - User requests should start with "The user's overall request for the assistant is to"
                    - Task descriptions should start with "The task is to"
                    - Make sure to omit any personally identifiable information (PII), like names, locations, phone numbers, email addressess, company names and so on.
                    - Make sure to indicate specific details such as programming languages, frameworks, libraries and so on which are relevant to the task. Also mention the specific source of the data if it's relevant to the task. For instance, if the user is asking to perform a calculation, let's talk about the specific source of the data.
                    """,
                }
            ],
            context={"messages": conversation.messages},
            response_model=ConversationSummary,
        )
        return {
            "chat_id": conversation.chat_id,
            "task_description": resp.task_description,
            "user_request": resp.user_request,
            "turns": len(conversation.messages),
        }



from openai import AsyncOpenAI
from asyncio import Semaphore
from helpers.types import Cluster


async def embed_summary(
    oai_client: AsyncOpenAI, sem: Semaphore, item: str, text_to_embed: str
) -> list[float]:
    return {
        "embedding": (
            await oai_client.embeddings.create(
                input=text_to_embed, model="text-embedding-3-small"
            )
        )
        .data[0]
        .embedding,
        **item,
    }


async def embed_cluster(
    oai_client: AsyncOpenAI, sem: Semaphore, cluster: Cluster
) -> list[float]:
    return {
        "embedding": (
            await oai_client.embeddings.create(
                input=f"""
                Name: {cluster.name}
                Description: {cluster.description}
                """,
                model="text-embedding-3-small",
            )
        )
        .data[0]
        .embedding,
        "cluster": cluster,
    }



import pandas as pd
from typing import List
from helpers.types import Conversation


def generate_cumulative_chart_data(conversations: List[Conversation]) -> dict:
    """
    Generate cumulative word count chart data for human messages in conversations.
    Returns a dict containing the Plotly data and layout.
    """
    messages_data = []
    for conv in conversations:
        for msg in conv.messages:
            if msg.role == "human":
                messages_data.append(
                    {
                        "datetime": pd.to_datetime(
                            str(msg.created_at).replace("Z", "+00:00")
                        ),
                        "words": len(msg.content.split()),
                    }
                )

    df = pd.DataFrame(messages_data)
    df["week_start"] = df["datetime"].dt.to_period("W-MON").dt.start_time

    weekly_df = df.groupby("week_start")["words"].sum().reset_index()
    weekly_df["cumulative_words"] = weekly_df["words"].cumsum()
    weekly_df["week_start"] = weekly_df["week_start"].dt.strftime("%Y-%m-%d")

    return [
        {"x": x, "y": y}
        for x, y in zip(
            weekly_df["week_start"].tolist(), weekly_df["cumulative_words"].tolist()
        )
    ]


def generate_messages_per_chat_data(conversations: List[Conversation]) -> dict:
    messages_data = []
    for conv in conversations:
        for msg in conv.messages:
            messages_data.append(
                {
                    "datetime": pd.to_datetime(
                        str(msg.created_at).replace("Z", "+00:00")
                    ),
                    "chat_id": conv.chat_id,
                }
            )

    df = pd.DataFrame(messages_data)
    df["week_start"] = df["datetime"].dt.to_period("W-MON").dt.start_time

    weekly_messages = df.groupby("week_start").size().reset_index(name="message_count")
    weekly_chats = (
        df.groupby("week_start")["chat_id"].nunique().reset_index(name="chat_count")
    )

    weekly_df = pd.merge(weekly_messages, weekly_chats, on="week_start")
    weekly_df["avg_messages"] = weekly_df["message_count"] / weekly_df["chat_count"]
    weekly_df["week_start"] = weekly_df["week_start"].dt.strftime("%Y-%m-%d")

    return [
        {"x": x, "y": y}
        for x, y in zip(
            weekly_df["week_start"].tolist(), weekly_df["avg_messages"].tolist()
        )
    ]


def generate_messages_per_week_data(conversations: List[Conversation]) -> dict:
    messages_data = []
    for conv in conversations:
        for msg in conv.messages:
            messages_data.append(
                {
                    "datetime": pd.to_datetime(
                        str(msg.created_at).replace("Z", "+00:00")
                    ),
                    "chat_id": conv.chat_id,
                }
            )

    df = pd.DataFrame(messages_data)
    df["week_start"] = df["datetime"].dt.to_period("W-MON").dt.start_time

    weekly_messages = df.groupby("week_start").size().reset_index(name="message_count")
    weekly_messages["week_start"] = weekly_messages["week_start"].dt.strftime(
        "%Y-%m-%d"
    )

    return [
        {"x": x, "y": y}
        for x, y in zip(
            weekly_messages["week_start"].tolist(),
            weekly_messages["message_count"].tolist(),
        )
    ]


def generate_new_chats_per_week_data(conversations: List[Conversation]) -> dict:
    chat_starts = pd.DataFrame(
        [
            {
                "datetime": pd.to_datetime(str(conv.created_at).replace("Z", "+00:00")),
                "chat_id": conv.chat_id,
            }
            for conv in conversations
        ]
    )
    chat_starts["week_start"] = (
        chat_starts["datetime"].dt.to_period("W-MON").dt.start_time
    )
    weekly_chats = (
        chat_starts.groupby("week_start").size().reset_index(name="chat_count")
    )
    weekly_chats["week_start"] = weekly_chats["week_start"].dt.strftime("%Y-%m-%d")

    return [
        {"x": x, "y": y}
        for x, y in zip(
            weekly_chats["week_start"].tolist(), weekly_chats["chat_count"].tolist()
        )
    ]



import numpy as np


def get_contrastive_examples(
    cluster_id: int,
    grouped_clusters: dict[int, list[dict]],
    desired_count: int = 10,
) -> dict[int, list[dict]]:
    other_clusters = [c for c in grouped_clusters.keys() if c != cluster_id]

    if len(other_clusters) == 0:
        raise ValueError("No other clusters to compare against")

    # Get all examples from other clusters
    all_examples = []
    for cluster in other_clusters:
        all_examples.extend(grouped_clusters[cluster])

    # If we don't have enough examples, use replacement
    if len(all_examples) < desired_count:
        return all_examples

    # Otherwise sample without replacement
    return list(np.random.choice(all_examples, size=desired_count, replace=False))



from fasthtml.common import (
    fast_app,
    Div,
    Titled,
    Link,
    Html,
    serve,
    H2,
    Form,
    Input,
    H1,
    Style,
    H3,
    Request,
    Script,
)
from helpers.types import Cluster
from helpers.conversation import load_conversations
from helpers.visualisation import (
    generate_cumulative_chart,
    generate_messages_per_chat_chart,
    generate_messages_per_week_chart,
    generate_new_chats_per_week_chart,
)
import os
from main import generate_clusters
import asyncio

app, rt = fast_app(
    hdrs=(
        Script(src="https://cdn.plot.ly/plotly-2.32.0.min.js"),
        Link(
            rel="stylesheet",
            href="https://cdnjs.cloudflare.com/ajax/libs/flexboxgrid/6.3.1/flexboxgrid.min.css",
            type="text/css",
        ),
    )
)


center_style = (
    Style(
        """
            body { 
                font-family: system-ui, sans-serif;
                max-width: 800px;
                margin: 2rem auto;
                padding: 0 1rem;
            }
            form {
                display: flex;
                flex-direction: column;
                gap: 1rem;
                padding: 2rem;
                border-radius: 8px;
                border: 1px solid #ccc;
            }
            input[type="submit"] {
                background: #0066ff;
                color: white;
                padding: 0.5rem 1rem;
                border: none;
                border-radius: 4px;
                cursor: pointer;
            }
            input[type="submit"]:hover {
                background: #0052cc;
            }
            .missing-file {
                background: #fff3cd;
                border: 1px solid #ffeeba;
                padding: 1rem;
                margin: 1rem 0;
                border-radius: 4px;
            }
            .loading {
                text-align: center;
                padding: 2rem;
                background: #f8f9fa;
                border-radius: 8px;
                box-shadow: 0 2px 4px rgba(0,0,0,0.1);
                margin: 2rem auto;
                max-width: 400px;
            }
            .loading-spinner {
                display: inline-block;
                width: 50px;
                height: 50px;
                border: 3px solid #f3f3f3;
                border-top: 3px solid #0066ff;
                border-radius: 50%;
                animation: spin 1s linear infinite;
                margin-bottom: 1rem;
            }
            @keyframes spin {
                0% { transform: rotate(0deg); }
                100% { transform: rotate(360deg); }
            }
            .submit-btn {
                display: inline-block;
                margin: 0.5rem;
                background: #0066ff;
                color: white;
                padding: 0.5rem 1rem;
                border: none;
                border-radius: 4px;
                cursor: pointer;
                font-size: 0.75rem;
            }
"""
    ),
)


@rt("/analyze")
async def analyze(request: Request):
    step = request.query_params.get("step")

    return Div(
        center_style,
        H1("Analyzing Your Chats", style="text-align: center;"),
        Div(
            Div(cls="loading-spinner"),
            Div(
                "Please wait while we analyze your conversations...",
                style="color: #666;",
            ),
            cls="loading",
        ),
        Script(
            f"""
            async function startAnalysis() {{
                try {{
                    await fetch('/start-analysis{f"?step={step}" if step else ""}');
                    window.location.href = '/';
                }} catch (error) {{
                    console.error('Analysis failed:', error);
                }}
            }}
            startAnalysis();
        """
        ),
    )


@rt("/start-analysis")
async def start_analysis(request: Request):
    step = request.query_params.get("step")
    await generate_clusters(start_step=step if step else None)
    return "Analysis complete"


def create_analysis_buttons():
    return Div(
        H3("Analysis Options", style="text-align: center;"),
        Div(
            Form(
                Input(
                    type="submit",
                    value="Start Fresh Analysis",
                    cls="submit-btn",
                ),
                method="POST",
                action="/analyze",
                style="display: inline-block;",
            ),
            *[
                Form(
                    Input(
                        type="submit",
                        value=step_label,
                        cls="submit-btn",
                    ),
                    method="POST",
                    action=f"/analyze?step={step_name}",
                    style="display: inline-block;",
                )
                for step_name, (step_label, prev_checkpoint) in {
                    "summarize": ("Re-summarize Conversations", "summaries.json"),
                    "embed": ("Re-embed Summaries", "embedded_summaries.json"),
                    "cluster": ("Re-cluster Summaries", "base_clusters.json"),
                }.items()
                if os.path.exists(f"checkpoints/{prev_checkpoint}")
            ],
            style="display: flex; flex-wrap: wrap; justify-content: center; gap: 1rem;",
        ),
        style="margin-bottom: 2rem; background: #f8f9fa; padding: 1rem; border-radius: 8px;",
    )


@rt("/")
async def index():
    if not os.path.exists("conversations.json"):
        return Div(
            center_style,
            H1("Chat Analysis Setup"),
            Div(
                "Please download your Anthropic chat history first", cls="missing-file"
            ),
        )

    conversations = load_conversations("conversations.json")

    # Replace both instances of the analysis buttons with the function call
    analysis_options = create_analysis_buttons()

    if not os.path.exists("clusters.json"):
        return Div(
            center_style,
            H1("Chat Analysis Setup"),
            analysis_options,
        )

    with open("clusters.json", "r") as f:
        clusters = [Cluster.model_validate_json(line) for line in f]

    # Create tree structure for visualization
    def build_tree(cluster, processed_clusters=None):
        if processed_clusters is None:
            processed_clusters = set()

        # Skip if we've already processed this cluster
        if cluster.id in processed_clusters:
            return None

        processed_clusters.add(cluster.id)

        # Create node for current cluster
        node_id = f"{cluster.name.lower()}-{cluster.count}".replace(" ", "-")

        # Find children first to determine if we need expand button
        children = [c for c in clusters if c.parent_id == cluster.id]

        inner_elements = [
            H2(cluster.name, style="font-size: 1.2rem; margin: 0.5rem 0;"),
            Div(f"Description: {cluster.description}", style="font-size: 0.9rem;"),
            Div(
                f"Contains {cluster.count} conversations",
                style="font-size: 0.9rem;",
            ),
        ]

        # Only add expand button if there are children
        if children:
            inner_elements.append(
                Div(
                    "Expand",
                    cls="expand-btn",
                    id=f"btn-{node_id}",
                    style="""
                        font-size: 0.8rem;
                        padding: 0.2rem 0.5rem;
                        background: #eee;
                        border-radius: 4px;
                        cursor: pointer;
                        display: inline-block;
                        margin-top: 0.5rem;
                    """,
                    onclick=f"""
                        const children = document.getElementById('children-{node_id}');
                        if (children) {{
                            if (children.style.display === 'none') {{
                                children.style.display = 'block';
                                this.textContent = 'Collapse';
                            }} else {{
                                children.style.display = 'none';
                                this.textContent = 'Expand';
                            }}
                        }}
                        event.stopPropagation();
                    """,
                )
            )

        node = Div(
            Div(
                *inner_elements,
                style="text-align: left;",
            ),
            cls="cluster-node",
            id=node_id,
            style="padding: 1rem; border: 1px solid #ccc; margin: 1rem; border-radius: 4px;",
        )

        # Add children recursively if they exist
        if children:
            child_nodes = [build_tree(child, processed_clusters) for child in children]
            # Filter out None values from already processed clusters
            child_nodes = [n for n in child_nodes if n is not None]
            if child_nodes:
                children_div = Div(
                    *child_nodes,
                    style="margin-left: 2rem; display: none;",  # Hidden by default
                    cls="children",
                    id=f"children-{node_id}",
                )
                return Div(node, children_div)
        return node

    # Find root clusters and build tree
    root_clusters = [c for c in clusters if not c.parent_id]
    cluster_tree = Div(
        H3(
            "Cluster Analysis",
            style="width: 100%; text-align: center; margin-bottom: 2rem;",
        ),
        Div(
            Div(
                *[build_tree(root) for root in root_clusters],
                style="background: white; color: black; padding: 2rem; max-height: 800px; overflow-y: auto;",
            ),
        ),
    )

    return Titled(
        "Chat Analysis",
        create_analysis_buttons(),  # Replace the duplicate buttons with function call
        # Existing visualization components
        Div(
            Div(
                Div(
                    cls="box",
                    id="myDiv",
                ),
                cls="col-xs-6 center-xs",
            ),
            Div(
                Div(
                    cls="box",
                    id="chatsPerWeekDiv",
                ),
                cls="col-xs-6 center-xs",
            ),
            Div(
                Div(
                    cls="box",
                    id="messagesDiv",
                ),
                cls="col-xs-6 center-xs",
            ),
            Div(
                Div(
                    cls="box",
                    id="messagesPerWeekDiv",
                ),
                cls="col-xs-6 center-xs",
            ),
            cls="row center-xs",
            style="color: #fff;",
        ),
        Style(
            """
            .cluster-node {
                background: #f5f5f5;
            }
            .cluster-node:hover {
                background: #e5e5e5;
            }
        """
        ),
        cluster_tree,
        generate_cumulative_chart(conversations),
        generate_messages_per_chat_chart(conversations),
        generate_new_chats_per_week_chart(conversations),
        generate_messages_per_week_chart(conversations),
        Div(
            style="padding: 40px;",
        ),
    )


serve()



import instructor
from asyncio import Semaphore
from helpers.embedding import embed_cluster, embed_summary
import math
from helpers.types import (
    CandidateClusters,
    Cluster,
    ClusterLabel,
    ClusterSummary,
    ConsolidatedCluster,
)
from helpers.sampling import get_contrastive_examples
from tqdm.asyncio import tqdm_asyncio as asyncio
from openai import AsyncOpenAI
from sklearn.cluster import KMeans
import numpy as np
import random
import os
import json


def cluster_items(items: list[dict], items_per_cluster: int) -> list[dict]:
    n_clusters = math.ceil(len(items) // items_per_cluster)
    embeddings = [item["embedding"] for item in items]
    X = np.array(embeddings)

    kmeans = KMeans(n_clusters=n_clusters)
    cluster_labels = kmeans.fit_predict(X)

    group_to_clusters = {}
    for i, item in enumerate(items):
        item_copy = item.copy()
        cluster_id = int(cluster_labels[i])
        del item_copy["embedding"]
        if cluster_id not in group_to_clusters:
            group_to_clusters[cluster_id] = []

        group_to_clusters[cluster_id].append(item_copy)

    return group_to_clusters


async def generate_base_cluster(
    client,
    sem,
    user_requests: list[dict],
    negative_examples: list[dict],
):
    async with sem:
        cluster = await client.chat.completions.create(
            messages=[
                {
                    "role": "system",
                    "content": """
                You are tasked with summarizing a group of related statements into a short, precise, and accurate description and name. Your goal is to create a concise summary that captures the essence of these statements and distinguishes them from other similar groups of statements.
                 
                Summarize all the statements into a clear, precise, two-sentence description in the past tense. Your summary should be specific to this group and distinguish it from the contrastive answers of the other groups.

                After creating the summary, generate a short name for the group of statements. This name should be at most ten words long (perhaps less) and be specific but also reflective of most of the statements (rather than reflecting only one or two). The name should distinguish this group from the contrastive examples. For instance, "Write fantasy sexual roleplay with octopi and monsters", "Generate blog spam for gambling websites", or "Assist with high school math homework" would be better and more actionable than general terms like "Write erotic content" or "Help with homework". Be as descriptive as possible and assume neither good nor bad faith. Do not hesitate to identify and describe socially harmful or sensitive topics specifically; specificity is necessary for monitoring
                 
                The cluster name should be a sentence in the imperative that captures the user’s request. For example, ‘Brainstorm ideas for a birthday party’ or ‘Help me find a new job.”
                """,
                },
                {
                    "role": "user",
                    "content": """
                Here are the relevant statements

                Below are the related statements:
                <positive_examples>
                {% for item in user_requests %}
                    <positive_example>{{ item['user_request'] }}</positive_example>
                {% endfor %}
                </positive_examples>  
                 
                For context, here are statements from nearby groups that are NOT part of the group you’re summarizing
                 
                <contrastive_examples>
                {% for item in negative_examples %}
                <contrastive_example>{{ item['user_request'] }}</contrastive_example>
                {% endfor %}
                </contrastive_examples>
                 
                Remember to analyze both the statements and the contrastive statements carefully to ensure your summary and name accurately represent the specific group while distinguishing it from others.
                """,
                },
            ],
            response_model=ClusterSummary,
            context={
                "user_requests": user_requests,
                "negative_examples": negative_examples,
            },
        )

        return Cluster(
            name=cluster.name,
            description=cluster.summary,
            chat_ids=[item["chat_id"] for item in user_requests],
            parent_id=None,
        )


async def embed_summaries(
    oai_client: AsyncOpenAI,
    sem: Semaphore,
    summaries: list[dict],
    output_file: str = "checkpoints/embedded_summaries.json",
    start_step: str = None,
) -> list[dict]:
    should_reembed = start_step in ["embed", "summarize"] if start_step else False

    if not should_reembed and os.path.exists(output_file):
        with open(output_file, "r") as f:
            return [json.loads(line) for line in f]

    embedded_summaries = await asyncio.gather(
        *[
            embed_summary(oai_client, sem, summary, summary["task_description"])
            for summary in summaries
        ],
        desc="Embedding summaries",
    )

    os.makedirs("checkpoints", exist_ok=True)
    with open(output_file, "w") as f:
        for summary in embedded_summaries:
            f.write(json.dumps(summary) + "\n")

    return embedded_summaries


async def cluster_summaries(
    client: instructor.Instructor,
    sem: Semaphore,
    summaries: list[dict],
    items_per_cluster: int = 5,
    start_step: str = None,
):
    oai_client = AsyncOpenAI()
    embedded_summaries = await embed_summaries(
        oai_client, sem, summaries, start_step=start_step
    )

    cluster_id_to_summaries = cluster_items(embedded_summaries, items_per_cluster)

    return await asyncio.gather(
        *[
            generate_base_cluster(
                client,
                sem,
                cluster_id_to_summaries[cluster_id],
                get_contrastive_examples(
                    cluster_id, cluster_id_to_summaries, desired_count=10
                ),
            )
            for cluster_id in cluster_id_to_summaries
        ]
    )


async def label_cluster(
    client: instructor.AsyncInstructor,
    sem: Semaphore,
    cluster: dict,
    candidate_clusters: list[str],
) -> dict:
    async with sem:
        candidate_clusters = candidate_clusters.copy()
        random.shuffle(candidate_clusters)
        label = await client.chat.completions.create(
            messages=[
                {
                    "role": "system",
                    "content": """
                    You are tasked with categorizing a specific cluster into one of the provided higher-level clusters for observability, monitoring, and content moderation. Your goal is to determine which higher-level cluster best fits the given specific cluster based on its name and description.

                    First, here are the ONLY valid higher-level clusters you may select from:
                    <higher_level_clusters>
                    {% for cluster in candidate_clusters %}
                    <higher_level_cluster>{{ cluster }}</higher_level_cluster>
                    {% endfor %}
                    </higher_level_clusters>

                    Here is the specific cluster to categorize:
                    <specific_cluster>
                    Name: {{ cluster["name"] }}
                    Description: {{ cluster["description"] }}
                    </specific_cluster>

                    RULES:
                    1. You MUST select EXACTLY ONE higher-level cluster from the provided list
                    2. You MUST output the higher-level cluster name EXACTLY as written - no modifications allowed
                    3. You MUST NOT create new cluster names or combinations
                    4. You MUST NOT output any additional text or explanations
                    5. You MUST NOT use partial matches or approximate names

                    CLASSIFICATION PROCESS:
                    1. First, record the exact list of valid higher-level clusters
                    2. Read the specific cluster's name and description carefully
                    3. Compare the specific cluster's key characteristics against each valid higher-level cluster
                    4. Select the single most appropriate higher-level cluster that encompasses the specific cluster
                    5. Verify your selected cluster exactly matches one from the valid list
                    6. Output ONLY the selected higher-level cluster name, exactly as it appeared in the valid list

                    Here is the specific cluster to categorize:

                    <specific_cluster>
                    Name: {{ cluster["name"] }}
                    Description: {{ cluster["description"] }}
                    </specific_cluster>
                    
                    Based on this information, determine the most appropriate higher-level cluster and provide your answer as instructed.
                    """,
                }
            ],
            context={
                "cluster": cluster,
                "candidate_clusters": candidate_clusters,
            },
            response_model=ClusterLabel,
            max_retries=3,
        )
    return {
        "cluster": cluster,
        "higher_level_cluster": label.higher_level_cluster,
    }


async def generate_candidate_cluster_names(
    client: instructor.Instructor, sem: Semaphore, clusters: list[Cluster]
) -> list[str]:
    resp = await client.chat.completions.create(
        messages=[
            {
                "role": "system",
                "content": """
                You are tasked with creating higher-level cluster names based on a given list of clusters and their descriptions. Your goal is to come up with broader categories that could encompass one or more of the provided clusters

                First, review the list of clusters and their descriptions:
                <cluster_list>
                    {% for cluster in clusters %}
                    <cluster>{{ cluster.name }}: {{ cluster.description }}</cluster>
                    {% endfor %}
                </cluster_list>

                Your task is to create at most {{ desired_number }} higher-level cluster names that could potentially include one or more of the provided clusters. These higher-level clusters should represent broader categories or themes that emerge from the given clusters, while remaining as specific as possible. If there are many clusters with a specific theme, ensure that the higher-level cluster name remains the maximum level of specificity. You are helping to organize user behavior data in order to improve safety, monitoring, and observability. You can generate less than {{ desired_number }} names if you feel that fewer are appropriate and accurately capture the clusters. 

                Guidelines for creating higher-level clusters names
                1. Analyze the themes, topics or characteristics common to multiple clusters.
                2. Create names that are specific enough to be meaningful but but not so specific that they can’t meaningfully represent many different clusters. Avoid overly general or vague terms, and do not hesitate to describe socially harmful or sensitive topics (in fact, clusters that clearly describe harmful behavior are slightly preferred); specificity is necessary for observability and enforcement.
                3. Ensure that the higher-level cluster names are distinct from one another.
                4. Use clear, concise, and descriptive language for the cluster names. Assume neither good nor bad faith for the content in the clusters.

                Think about the relationships between the given clusters and potential overarching themes. 
                
                Focus on creating meaningful, distinct and precise ( but not overly specific ) higher-level cluster names that could encompass multiple sub-clusters.
                """,
            }
        ],
        response_model=CandidateClusters,
        context={
            "clusters": clusters,
            "desired_number": math.ceil(
                (len(clusters) * 3) // 4
            ),  # This ensure we get at least 3/4 of the clusters ( so we have a 25% reduction or more at each stage)
        },
        max_retries=3,
    )

    return resp.candidate_cluster_names


async def rename_higher_level_cluster(
    client: instructor.AsyncInstructor,
    sem: Semaphore,
    clusters: dict,
) -> dict:
    async with sem:
        resp = await client.chat.completions.create(
            messages=[
                {
                    "role": "system",
                    "content": """
                    You are tasked with summarizing a group of related cluster names into a short, precise, and accurate overall description and name. Your goal is to create a concise summary that captures the essence of these clusters

                    The cluster name should be at most ten words long (perhaps less) and be specific but also reflective of most of the clusters that comprise them. The cluster name should be a sentence in the imperative that captures the user’s request. For example, ‘Brainstorm ideas for a birthday party’ or ‘Help me find a new job are good examples.

                    Do not shy away from describing socially harmful or sensitive topics. For instance, "Write fantasy sexual roleplay with octopi and monsters", "Generate blog spam for gambling websites", or "Assist with high school math homework" would be better than "Write erotic content" or "Help with homework". Be as descriptive as possible and assume neither good nor bad faith. Do not hesitate to identify and describe socially harmful or sensitive topics specifically; specificity is necessary for monitoring.

                    The cluster description should be a clear, precise, two-sentence description in the past tense. This description should be specific to this cluster.    
                    
                    Below are the related cluster names
                    <clusters>
                        {% for cluster in clusters %}
                            <cluster>{{ cluster.name }}: {{ cluster.description }}</cluster>
                        {% endfor %}
                    </clusters>

                    Ensure your summary and name accurately represent the clusters and are specific to the clusters.
                    """,
                }
            ],
            context={"clusters": clusters},
            response_model=ConsolidatedCluster,
        )

    return {
        "new_cluster": Cluster(
            name=resp.cluster_name,
            description=resp.cluster_description,
            chat_ids=[chat_id for cluster in clusters for chat_id in cluster.chat_ids],
            parent_id=None,
        ),
        "original_clusters": clusters,
    }


async def generate_higher_level_cluster(
    client: instructor.Instructor,
    sem: Semaphore,
    clusters: list[Cluster],
) -> list[Cluster]:
    if len(clusters) == 1:
        original_cluster = clusters[0]
        new_cluster = Cluster(
            name=original_cluster.name,
            description=original_cluster.description,
            chat_ids=original_cluster.chat_ids,
            parent_id=None,
        )
        original_cluster.parent_id = new_cluster.id
        return [original_cluster, new_cluster]

    candidate_cluster_names = await generate_candidate_cluster_names(
        client, sem, clusters
    )

    new_candidate_clusters = await asyncio.gather(
        *[
            label_cluster(client, sem, cluster, candidate_cluster_names)
            for cluster in clusters
        ],
        desc="Generating candidate higher-level clusters",
    )

    cluster_groups: dict[str, list[Cluster]] = {}
    for cluster in new_candidate_clusters:
        original_cluster = cluster["cluster"]
        new_higher_level_cluster = cluster["higher_level_cluster"]
        if new_higher_level_cluster not in cluster_groups:
            cluster_groups[new_higher_level_cluster] = []
        cluster_groups[new_higher_level_cluster].append(original_cluster)

    new_labelled_clusters = await asyncio.gather(
        *[
            rename_higher_level_cluster(client, sem, cluster_group)
            for cluster_group in cluster_groups.values()
        ],
        desc="Generating higher-level clusters",
    )

    res = []

    for new_labelled_cluster in new_labelled_clusters:
        res.append(new_labelled_cluster["new_cluster"])

        for original_cluster in new_labelled_cluster["original_clusters"]:
            original_cluster.parent_id = new_labelled_cluster["new_cluster"].id
            res.append(original_cluster)

    return res


async def reduce_clusters(
    client: instructor.Instructor,
    sem: Semaphore,
    clusters: list[Cluster],
    items_per_cluster: int,
) -> list[Cluster]:
    oai_client = AsyncOpenAI()
    embedded_clusters = await asyncio.gather(
        *[embed_cluster(oai_client, sem, cluster) for cluster in clusters]
    )

    cluster_id_to_cluster = {
        cluster_id: [item["cluster"] for item in clusters]
        for cluster_id, clusters in cluster_items(
            embedded_clusters, items_per_cluster
        ).items()
    }

    new_clusters = await asyncio.gather(
        *[
            generate_higher_level_cluster(
                client,
                sem,
                cluster_id_to_cluster[cluster_id],
            )
            for cluster_id in cluster_id_to_cluster
        ],
        desc="Generating higher-level clusters",
    )
    return [
        cluster
        for clusters in new_clusters
        if clusters is not None
        for cluster in clusters
    ]



from helpers.types import Cluster
from rich import print
from rich.tree import Tree as RichTree

if __name__ == "__main__":
    with open("clusters.json", "r") as f:
        clusters = [Cluster.model_validate_json(line) for line in f]

    # Create a rich tree for better visualization
    rich_tree = RichTree("Clusters")

    # First create a mapping of cluster IDs to their objects
    cluster_map = {cluster.id: cluster for cluster in clusters}

    # Find root clusters (those without parents)
    root_clusters = [c for c in clusters if not c.parent_id]

    # Keep track of processed clusters to avoid duplicates
    processed_clusters = set()

    def add_children(parent_tree, parent_cluster):
        # Find all children of this cluster
        children = [c for c in clusters if c.parent_id == parent_cluster.id]
        for child in children:
            # Skip if we've already processed this cluster
            if child.id in processed_clusters:
                continue
            processed_clusters.add(child.id)

            # Create a node with cluster info
            child_node = parent_tree.add(
                f"[bold]{child.name}[/bold]\n"
                f"Description: {child.description}\n"
                f"Contains {child.count} conversations"
            )
            # Recursively add children
            add_children(child_node, child)

    # Add all root clusters and their children
    for root in root_clusters:
        if root.id not in processed_clusters:
            processed_clusters.add(root.id)
            root_node = rich_tree.add(
                f"[bold blue]{root.name}[/bold blue]\n"
                f"Description: {root.description}\n"
                f"Contains {root.count} conversations"
            )
            add_children(root_node, root)

    print("Combined View")
    print(rich_tree)



from fastapi import FastAPI, staticfiles
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel
from pathlib import Path
from kura.types import Cluster, Conversation, Message
from kura.cli.visualisation import (
    generate_cumulative_chart_data,
    generate_messages_per_chat_data,
    generate_messages_per_week_data,
    generate_new_chats_per_week_data,
)
import json

api = FastAPI()

# Configure CORS
api.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],  # Allows all origins
    allow_credentials=True,
    allow_methods=["*"],  # Allows all methods
    allow_headers=["*"],  # Allows all headers
)

# Serve static files from web/dist
web_dir = Path(__file__).parent.parent / "static" / "dist"
if not web_dir.exists():
    raise FileNotFoundError(f"Static files directory not found: {web_dir}")


class ConversationData(BaseModel):
    data: list[dict]


@api.post("/api/analyse")
async def analyse_conversations(conversation_data: ConversationData):
    conversations = [
        Conversation(
            chat_id=conversation["uuid"],
            created_at=conversation["created_at"],
            messages=[
                Message(
                    created_at=message["created_at"],
                    role=message["sender"],
                    content="\n".join(
                        [
                            item["text"]
                            for item in message["content"]
                            if item["type"] == "text"
                        ]
                    ),
                )
                for message in conversation["chat_messages"]
            ],
        )
        for conversation in conversation_data.data
    ]

    # Load clusters from checkpoint file if it exists
    clusters_file = Path(__file__).parent.parent.parent / "checkpoints/clusters.json"
    clusters = []
    print(clusters_file)
    if clusters_file.exists():
        print("")
        with open(clusters_file) as f:
            clusters_data = []
            for line in f:
                clusters_data.append(line)
            clusters = [Cluster(**json.loads(cluster)) for cluster in clusters_data]

    return {
        "cumulative_words": generate_cumulative_chart_data(conversations),
        "messages_per_chat": generate_messages_per_chat_data(conversations),
        "messages_per_week": generate_messages_per_week_data(conversations),
        "new_chats_per_week": generate_new_chats_per_week_data(conversations),
        "clusters": clusters,
    }


# Serve static files from web/dist at the root
web_dir = Path(__file__).parent.parent / "static" / "dist"
if not web_dir.exists():
    raise FileNotFoundError(f"Static files directory not found: {web_dir}")

# Mount static files at root
api.mount("/", staticfiles.StaticFiles(directory=str(web_dir), html=True))



import typer
import uvicorn
from kura.cli.server import api
from rich import print

app = typer.Typer()


@app.command()
def start_app():
    """Start the FastAPI server"""

    uvicorn.run(api, host="0.0.0.0", port=8000)
    print(
        "\n[bold green]🚀 Access website at[/bold green] [bold blue][http://localhost:8000](http://localhost:8000)[/bold blue]\n"
    )


if __name__ == "__main__":
    app()



from .summarise import SummariseBase
from .summary_cluster import SummaryCluster
from .types import Cluster, Conversation, Message
import json
import os
from .cluster import ClusterModel


class LLMCluster:
    def __init__(
        self,
        summariser: SummariseBase = SummariseBase(),
        base_cluster_model: SummaryCluster = SummaryCluster(),
        cluster_model: ClusterModel = ClusterModel(),
        max_clusters: int = 10,
        checkpoint_dir: str = "checkpoints",
    ):
        self.summariser = summariser
        self.base_cluster_model = base_cluster_model
        self.max_clusters = max_clusters
        self.checkpoint_dir = checkpoint_dir
        self.cluster_model = cluster_model
        
        self.clusters = []

        if not os.path.exists(self.checkpoint_dir):
            os.makedirs(self.checkpoint_dir)

    async def reduce_clusters(
        self, clusters: list[Cluster], root_clusters: list[Cluster], iteration: int = 1
    ):
        while len(root_clusters) > self.max_clusters:
            print(f"\n=== Iteration {iteration} ===")
            print(f"    Starting with {len(root_clusters)} root clusters...")

            # Generate new parent clusters from current level
            new_parent_clusters = await self.cluster_model.cluster_clusters(
                root_clusters
            )

            # This is the new layer
            root_clusters = [c for c in new_parent_clusters if not c.parent_id]

            # Remove the outdated versions of clusters that were just reduced
            old_cluster_ids = {rc.id for rc in new_parent_clusters if rc.parent_id}
            clusters = [c for c in clusters if c.id not in old_cluster_ids]

            # Add both the updated original clusters and new parent clusters
            clusters.extend(new_parent_clusters)
            print(f"    Successfully reduced to {len(root_clusters)} root clusters")

            iteration += 1

        return clusters, root_clusters

    async def cluster_conversations(self):
        summarized_conversations = await self.summariser.summarise_conversations(
            self.conversations
        )

        base_clusters = await self.base_cluster_model.cluster_conversation_summaries(
            summarized_conversations
        )

        clusters_path = os.path.join(self.checkpoint_dir, "clusters.json")

        if os.path.exists(clusters_path):
            with open(clusters_path, "r") as f:
                clusters = [Cluster.model_validate_json(line) for line in f]
        else:
            clusters = base_clusters
            root_clusters = base_clusters

            clusters, root_clusters = await self.reduce_clusters(
                clusters, root_clusters
            )

            # Save final clusters to checkpoint directory

            with open(clusters_path, "w") as f:
                for cluster in clusters:
                    f.write(cluster.model_dump_json() + "\n")

            
        

        print(f"\nSaved {len(clusters)} total clusters to {clusters_path}")
        
        return clusters

    def load_claude_messages(
        self, path: str, max_conversations: int = -1
    ) -> list[Conversation]:
        with open(path, "r") as f:
            conversations = []
            for conversation in json.load(f):
                conversations.append(
                    Conversation(
                        chat_id=conversation["uuid"],
                        created_at=conversation["created_at"],
                        messages=[
                            Message(
                                created_at=message["created_at"],
                                role=message["sender"],
                                content="\n".join(
                                    [
                                        item["text"]
                                        for item in message["content"]
                                        if item["type"] == "text"
                                    ]
                                ),
                            )
                            for message in conversation["chat_messages"]
                        ],
                    )
                )
            self.conversations = (
                conversations
                if max_conversations == -1
                else conversations[:max_conversations]
            )
            print(f"Read in {len(self.conversations)} conversations")



from pydantic import BaseModel, field_validator, computed_field, Field, ValidationInfo
import uuid


class Message(BaseModel):
    created_at: str
    role: str
    content: str


class Conversation(BaseModel):
    chat_id: str
    created_at: str
    messages: list[Message]


class Cluster(BaseModel):
    id: str = Field(
        default_factory=lambda: uuid.uuid4().hex,
    )
    name: str
    description: str
    parent_id: str | None
    chat_ids: list[str]

    @computed_field
    def count(self) -> int:
        return len(self.chat_ids)

    def add_child_count(self, child: "Cluster"):
        return Cluster(
            name=self.name,
            description=self.description,
            chat_ids=self.chat_ids + child.chat_ids,
        )


class ConversationSummary(BaseModel):
    task_description: str
    user_request: str

    @field_validator("user_request", "task_description")
    def validate_length(cls, v):
        if len(v.split()) > 25:
            raise ValueError(
                f"Please summarize the sentence {v} to at most 20 words. It's currently {len(v.split())} words long. Remove any unnecessary words and make sure to retain the most important information."
            )
        return v


class ClusterSummary(BaseModel):
    """
    This is a summary of the related statements into a single cluster that captures the essence of these statements and distinguishes them from the contrastive answers of other groups.

    Your job is to create a final name and summary for the cluster
    """

    name: str
    summary: str


class CandidateClusters(BaseModel):
    candidate_cluster_names: list[str]

    @field_validator("candidate_cluster_names")
    def validate_length(cls, v, info: ValidationInfo):
        desired_number = info.context["desired_number"]
        if len(v) > desired_number:
            raise ValueError(
                f"Too many candidate cluster names: got {len(v)}, expected at most {desired_number}. Please combine similar clusters to reduce the total number."
            )
        return v


class ClusterLabel(BaseModel):
    higher_level_cluster: str

    @field_validator("higher_level_cluster")
    def validate_length(cls, v, info: ValidationInfo):
        candidate_clusters = info.context["candidate_clusters"]
        if v not in candidate_clusters:
            raise ValueError(
                f"The chosen label of {v} is not in the list of candidate clusters provided of {candidate_clusters}. You must choose a label that belongs to the list of candidate clusters - {candidate_clusters}."
            )
        return v


class ConsolidatedCluster(BaseModel):
    """
    This is a consolidated cluster that is the result of merging multiple clusters into a single higher levle cluster
    """

    cluster_name: str
    cluster_description: str
